import datetime
import logging
import math

from collections import Counter

from odoo import api, fields, models, _
from odoo.exceptions import UserError, ValidationError

_logger = logging.getLogger(__name__)

class SaleSubscription(models.Model):
    _inherit = 'sale.order'

    maintenance_subscription_id = fields.Many2one('sale.order', copy=False, tracking=True)
    maintenance_main_subscription_id = fields.One2many("sale.order", "maintenance_subscription_id",
        help="Link to the main subscription for which this maintenance subscription applies")
    maintenance_is_activated = fields.Boolean(compute="_compute_maintenance_status",
        search="_search_maintenance_is_activated", string="Has Maintenance")
    maintenance_is_paying = fields.Boolean(compute="_compute_maintenance_status",
        search="_search_maintenance_is_paying", string="Paying Maintenance")
    maintenance_stage_category = fields.Selection(related="maintenance_subscription_id.subscription_state",
        string="Open Maintenance Subscription")

    _sql_constraints = [
        ('unique_maintenance_subscription', 'unique(maintenance_subscription_id)',
         'Only one subscription can be linked to a maintenance subscription'),
    ]

    @api.constrains('maintenance_main_subscription_id', 'subscription_state')
    def _check_open_main_maintenance(self):
        for rec in self:
            if (rec.subscription_state in ['3_progress', '4_paused'] and rec.maintenance_main_subscription_id
                    and rec.maintenance_main_subscription_id.subscription_state not in ('3_progress', '4_paused')):
                raise ValidationError(_("Maintenance activation is not allowed on closed subscriptions"))

    # computed field
    def _compute_maintenance_status(self):
        for rec in self:
            sub = rec._get_maintenance_subscription()
            maintenance_line = sub._get_maintenance_line()
            in_progress = sub.subscription_state in ['3_progress', '4_paused']
            rec.maintenance_is_activated = maintenance_line and in_progress
            rec.maintenance_is_paying = maintenance_line.price_subtotal > 0 and in_progress

    def _search_maintenance_is_activated(self, operator, value):
        maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        domain = ['|',
            '&',
                ('order_line.product_id', '=', maintenance_product.id),
                ('subscription_state', 'in', ('3_progress', '4_paused')),
            '&',
                ('maintenance_subscription_id.order_line.product_id', '=', maintenance_product.id),
                ('maintenance_stage_category', '=', ('3_progress', '4_paused')),
        ]
        if (operator, value) in [('=', True), ('!=', False)]:
            return domain
        # Not possibie to directly express, all the line should not have the product maintenance
        return [('id', 'not in', self.search(domain).ids)]

    def _search_maintenance_is_paying(self, operator, value):
        if (operator, value) in [('=', True), ('!=', False)]:
            return ['|',
                ('order_line.maintenance_is_paying', '=', True),
                ('maintenance_subscription_id.order_line.maintenance_is_paying', '=', True)
            ]
        return ['|',
            '&',
                ('maintenance_subscription_id', '=', False),
                ('order_line.maintenance_is_paying', '=', False),
            ('maintenance_subscription_id.order_line.maintenance_is_paying', '=', False)
        ]

    def action_open_linked_task(self):
        self.ensure_one()

        tasks = self.env['project.task'].search([('mnt_subscription_id', '=', self.id)])
        result = {
            "type": "ir.actions.act_window",
            "res_model": "project.task",
            "domain": [('id', 'in', tasks.ids)],
            "context": {
                "create": False,
                'search_default_group_project': 1,
                'search_default_is_open': 1,
            },
            "name": _("Subscription Tickets"),
            "view_mode": 'kanban,tree,form',
            "search_view_id": [self.env.ref('openerp_enterprise.task_by_subscription').id, 'search'],
        }

        return result

    # Maintenance Activation
    def _maintenance_line_description(self):
        desc = "Maintenance on {0} extra lines of code".format(self._get_maintenance_loc_count())
        if self.maintenance_main_subscription_id:
            desc += " On {}".format(self.maintenance_main_subscription_id.name)

        sub = self.maintenance_main_subscription_id or self
        maintenance_modules = sub.database_ids and sub.database_ids[0].maintenance_modules or _("None")
        desc += "\nCovered Extra Modules: {}".format(maintenance_modules)
        return desc

    def _maintenance_activate(self, rate, discount=0.0):
        self.ensure_one()
        if self.maintenance_is_activated:
            raise UserError(_("Maintenance already activated on this subscription"))

        maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        yearly = self.plan_id.billing_period_unit == 'year'
        qty = self._get_maintenance_quantity() if not yearly else 0
        if self.subscription_state == '7_upsell':
            qty = self.subscription_id._get_maintenance_quantity()
        if self.maintenance_subscription_id:
            maintenance_sub = self.maintenance_subscription_id
            maintenance_sub.action_confirm()
            maintenance_sub._get_maintenance_line().unlink()
        else:
            maintenance_sub = self

        line = {
            'product_id': maintenance_product.id,
            'product_uom_qty': qty,
            'product_uom': self.sale_order_template_id.get_product_uom(maintenance_product).id,
            'price_unit': rate,
            'discount': discount,
            'name': maintenance_sub._maintenance_line_description(),
        }
        maintenance_sub.write({'order_line': [(0, 0, line)]})
        return maintenance_sub

    def _move_info_to_renewing_subscription(self):
        """
            Link the maintenance subscription to the new confirmed renewed Sale order
            Link the new maintenance subscription to the main when renewed
        """
        super()._move_info_to_renewing_subscription()
        for sub in self:
            parent = sub.subscription_id
            if parent.maintenance_subscription_id:
                mnt_sub = parent.maintenance_subscription_id
                parent.maintenance_subscription_id = False
                # Flush to avoid triggering the unique constraint on maintenance_subscription_id
                parent.flush_recordset(['maintenance_subscription_id'])
                sub.maintenance_subscription_id = mnt_sub
            if parent.maintenance_main_subscription_id:
                parent.maintenance_main_subscription_id.maintenance_subscription_id = sub

    def action_activate_maintenance(self):
        self.ensure_one()
        if not self.env.su and not self.env.user.has_group('sales_team.group_sale_salesman'):
            raise UserError(_("You are not allowed to activate the maintenance"))

        rate = self._get_default_maintenance_rate()
        self.sudo()._maintenance_activate(rate)

    def action_open_maintenance_subscription(self):
        self.ensure_one()
        res = self.env['ir.actions.act_window']._for_xml_id('openerp_enterprise.action_oe_maintenance_subscription')
        res['res_id'] = self.maintenance_subscription_id.id or self.maintenance_main_subscription_id.id
        return res

    def _update_maintenance(self):
        for rec in self:
            line = rec._get_maintenance_line()
            # Only Update actual qty for monthly
            if not (rec.state == 'sale' and rec.locked) and line and rec.plan_id.billing_period_unit == 'month':
                line.write({
                    'product_uom_qty': rec._get_maintenance_quantity(),
                    'name': rec._maintenance_line_description(),
                })

    # Upsell
    def _maintenance_need_upsell(self):
        if not self or not self._get_maintenance_line() or self.plan_id.billing_period_unit == 'month':
            return False

        if self.subscription_state not in ['3_progress', '4_paused']:
            return False

        sold_qty = self._get_maintenance_line().product_uom_qty
        actual_qty = self._get_maintenance_quantity()
        if actual_qty <= sold_qty:
            return False

        # To avoid too frequent upsell, only need to upsell if an upsell
        # happened more than 30 days
        maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        last_month = datetime.date.today() - datetime.timedelta(days=30)
        upsell_done_count = self.env['sale.order.line'].search_count([
            ('order_id.subscription_id', '=', self.id),
            ('product_id', '=', maintenance_product.id),
            ('product_uom_qty', '>', 0),
            ('order_id.subscription_state', '=', '7_upsell'),
            ('order_id.state', '=', 'sale'),
            ('order_id.date_order', '>', last_month),
        ])

        return not upsell_done_count


    def _get_invoice_diff(self):
        res = super()._get_invoice_diff()
        maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        for subscription in self.filtered(lambda s: s._get_maintenance_line()):
            mnt_line = subscription._get_maintenance_line()
            sold_qty = mnt_line.product_uom_qty
            actual_qty = self._get_maintenance_quantity()
            upsell_lines = subscription.upsell_order_id.order_line.filtered(lambda l: l.product_id == maintenance_product)
            upsell_count = sum(l.product_uom_qty for l in upsell_lines)
            if (sold_qty + upsell_count) < actual_qty:
                mnt_upsell_data = {
                    'product_uom_qty': actual_qty - (sold_qty + upsell_count),
                    'product_id': mnt_line.product_id.id,
                    'product_uom': mnt_line.product_uom.id,
                    'name': mnt_line.name,
                }
                upsell = res.get(subscription.id)
                if upsell:
                    upsell['upsell'].append(mnt_upsell_data)
                else:
                    res[subscription.id] = {
                        'database_id': subscription.database_ids[:1].id,
                        'upsell': [mnt_upsell_data],
                    }
        return res

    def _sync_maintenance_subscription(self, vals):
        to_sync_field = ['referrer_id', 'commission_plan_id', 'partner_id']
        maintenance_subs = self.mapped('maintenance_subscription_id')
        sync_vals = {f: vals[f] for f in to_sync_field if f in vals}
        if maintenance_subs and sync_vals:
            maintenance_subs.write(sync_vals)

    def write(self, vals):
        res = super().write(vals)
        self._sync_maintenance_subscription(vals)
        return res

    def set_apps_recurring_lines(self):
        res = super().set_apps_recurring_lines()
        self._update_maintenance()
        return res

    # Helpers for qweb template
    def _get_maintenance_subscription(self):
        # Return the value of field maintenance_subscription_id, self otherwise
        # because in case of monthly subscription the maintenance line is directly on it
        return self.maintenance_subscription_id or self

    def _get_maintenance_line(self, check_maintenance_subscription=False):
        """
            Find the maintenance line on the subscription
            check_maintenance_subscription: True to check also on the maintenance_subscription_id
        """
        maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        subscription = self._get_maintenance_subscription() if check_maintenance_subscription else self
        lines = subscription.order_line.filtered(lambda x: x.product_id == maintenance_product)
        return lines and lines[0]

    def _get_maintenance_loc_count(self):
        subscription = self.maintenance_main_subscription_id or self
        if not subscription.database_ids:
            return 0

        return math.ceil(subscription.database_ids[0].maintenance_lines_count / 100.0) * 100

    def _get_maintenance_quantity(self):
        return self._get_maintenance_loc_count() / 100

    def _user_is_referrer(self, user=False):
        user = user or self.env.user
        return self.referrer_id.commercial_partner_id == user.commercial_partner_id

    def _get_default_maintenance_rate(self):
        maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        uom = self.sale_order_template_id.get_product_uom(maintenance_product)
        return maintenance_product.uom_id._compute_price(self.pricelist_id._get_product_price(maintenance_product, 1), uom)

    def _get_default_maintenance_cost(self):
        """
            Return the cost based on the loc
        """
        maintenance_line = self._get_maintenance_line(check_maintenance_subscription=True)
        if maintenance_line:
            rate = maintenance_line.price_unit
        else:
            rate = self._get_default_maintenance_rate()
        return rate * self._get_maintenance_quantity()

    def _get_next_maintenance_invoice_date(self):
        self.ensure_one()
        if self.maintenance_subscription_id:
            return self.maintenance_subscription_id.next_invoice_date
        else:
            return self.next_invoice_date

    def _get_tags_as_labels(self):
        """
            :return: [(label, classname), ...]
            based on the category of the products defined in the subscription
        """
        tags = {
            self.env.ref('openerp_enterprise.category_maintenance'): (_('Maintenance'), 'text-bg-info'),
            self.env.ref('openerp_enterprise.product_category_odoo_sh'): (_('Odoo.sh'), 'text-bg-danger'),
            self.env.ref('openerp_enterprise.category_self_subscription'): (_('Enterprise'), 'text-bg-warning'),
            self.env.ref('openerp_enterprise.product_category_odoo_users'): (_('Enterprise'), 'text-bg-warning'),
            self.env.ref('openerp_enterprise.product_category_partnership'): (_('Partnership'), 'text-bg-success'),
        }
        tag_name = set()
        for line in self.order_line:
            if tags.get(line.product_id.categ_id):
                tag_name.add(tags[line.product_id.categ_id])
        return list(tag_name)


class SaleSubscriptionLine(models.Model):
    _inherit = 'sale.order.line'

    maintenance_is_paying = fields.Boolean(compute="_compute_maintenance_is_paying", search="_search_maintenance_is_paying")

    def _compute_maintenance_is_paying(self):
        maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        for rec in self:
            rec.maintenance_is_paying = rec.product_id == maintenance_product and rec.price_subtotal > 0

    def _search_maintenance_is_paying(self, operator, value):
        maintenance_product = self.env.ref('openerp_enterprise.product_maintenance')
        if (operator, value) in [('=', True), ('!=', False)]:
            return ['&', ('product_id', '=', maintenance_product.id), ('price_subtotal', '>', 0)]
        return ['&', ('product_id', '=', maintenance_product.id), ('price_subtotal', '=', 0)]

    @api.constrains('order_id', 'product_id')
    def _check_one_maintenance_line(self):
        maintenance_product_id = self.env.ref('openerp_enterprise.product_maintenance')

        # filter out the non maintenance lines
        maintenance_lines = self.filtered(lambda x: x.product_id == maintenance_product_id)

        # Look for all maintenance lines on current subscription
        # Cannot have more line then subscription
        all_maintenance_lines = maintenance_lines.mapped('order_id.order_line').filtered(
            lambda x: x.product_id == maintenance_product_id)
        maintenance_line_counter = Counter(l.order_id for l in all_maintenance_lines)
        if maintenance_line_counter and max(maintenance_line_counter.values()) > 1:
            raise UserError(_("Maintenance subscriptions should contain only one maintenance line"))
